from __future__ import annotations

from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    NoReturn,
    Optional,
    Tuple,
    Type,
    Union,
)
import warnings

from makefun import partial

from open_mafia_engine.core.all import ABILITY, EActivate, Game, GameBuilder, get_path
from open_mafia_engine.util.matcher import FuzzyMatcher

from .lobby import AbstractLobby, SimpleDictLobby
from .parser import AbstractCommandParser, ShellCommandParser
from .raw import RawCommand, TUser, MafiaBadCommand

__all__ = ["command", "CommandHandler", "CommandRunner"]


class CommandHandler(object):
    """Descriptor, wrapper for command functions.

    Parameters
    ----------
    func : Callable
        The wrapped function.
    game : bool
        If True, allows use during the game. Default is True.
    lobby : bool
        If True, allows use in the lobby. Default is False.
    admin : bool
        If True, this is an admin-only command.
    name : None or str
        Command name. If None (default), uses function's name.
    """

    registered: Dict[str, CommandHandler] = {}
    score_cutoff: int = 80

    def __init__(
        self,
        func: Callable,
        *,
        game: bool = True,
        lobby: bool = False,
        admin: bool = False,
        name: str = None,
    ):
        if not callable(func):
            raise TypeError(f"`func` must be callable, got {func!r}")
        if name is None:
            name = func.__name__
        # TODO: Get documentation :)

        self.func = func
        self.name = name
        self.game = bool(game)
        self.lobby = bool(lobby)
        self.admin = bool(admin)

    def __set_name__(self, owner: Type[CommandRunner], name_: str):
        self.bound_name = name_
        _RC = "registered_commands"

        # self.registered[name] = self  # FIXME: change to the following?

        # Create a dict for `owner`
        if not hasattr(owner, _RC):
            v = dict(**getattr(super(owner), _RC, {}))
            setattr(owner, _RC, v)
        owner.registered_commands[self.name] = self

    def __get__(self, obj: CommandRunner, objtype=None):
        return partial(self.func, obj)


def command(
    name_or_func: Union[str, Callable],
    *,
    game: bool = None,
    lobby: bool = None,
    admin: bool = False,
) -> Union[Callable]:
    """Decorator to register methods as commands.

    By default, "game" is True, "lobby" is `not game`, and "admin" is False.
    """

    if game is None:
        if lobby is None:
            game = True
            lobby = False
        else:
            game = not lobby
    else:
        if lobby is None:
            lobby = not game
    game = bool(game)
    lobby = bool(lobby)
    admin = bool(admin)

    if isinstance(name_or_func, str):

        def inner(func: Callable) -> CommandHandler:
            return CommandHandler(
                func, name=name_or_func, game=game, lobby=lobby, admin=admin
            )

        return inner
    elif callable(name_or_func):
        func = name_or_func
        return CommandHandler(func, game=game, lobby=lobby, admin=admin)
    else:
        raise TypeError(f"Argument to `command` is neither str, nor callable.")


class CommandRunner(Generic[TUser]):
    """Interprets and runs commands on the Mafia engine.

    This handles both pre-game commands (such as joining or starting the game)
    and commands during the game.
    Basic Mafia commands are already added.

    To add or override a command handler, use the `command` decorator:

        class MyRunner(CommandRunner[MyUserType]):
            @command("foo", game=True)
            def my_foo(self, source: str, *args, **kwargs):
                return

    The command decorator defaults to non-admin, in-game actions.
    Specifying `lobby=True` will, by default, allow only lobby use.
    If you want it always on, use `@command("name", game=True, lobby=True)`

    Commands that don't match any given command will call `default_command()`.
    The default implementation raises a MafiaBadCommand exception, but you may
    override it (e.g. to ignore it).

    Attributes
    ----------
    parser : AbstractCommandParser
        Your own parser implementation, or ShellCommandParser() by default.
    lobby : AbstractLobby
        Your own lobby implementation, or SimpleDictLobby[TUser]() by default.
    game : None or Game
        The game state. Defaults to None.
    score_cutoff : int
        The score cutoff for matching command names, from 0 to 100.
        Default is 80 (relatively strict).
    """

    registered_commands: Dict[str, CommandHandler]

    def __init__(
        self,
        parser: AbstractCommandParser = ShellCommandParser(),
        lobby: AbstractLobby[TUser] = SimpleDictLobby[TUser](),
        game: Optional[Game] = None,
        *,
        score_cutoff: int = 80,
    ):
        self.parser = parser
        self.lobby = lobby
        self.score_cutoff = int(score_cutoff)
        # This will check :)
        self.game = game

    @classmethod
    def all_available_commands(cls) -> List[str]:
        """Returns all commands that are registered."""
        return list(cls.registered_commands.keys())

    def currently_available_commands(self, source: str = None) -> List[str]:
        """Returns all commands that are currently available."""
        all_cmds = self.all_available_commands()

        def chk(cmd: str) -> bool:
            # FIXME
            try:
                self.pre_dispatch_check(cmd, source=source)
            except Exception:
                return False
            return True

        return [cmd for cmd in all_cmds if chk(cmd)]

    @property
    def game(self) -> Optional[Game]:
        return self._game

    @game.setter
    def game(self, game: Optional[Game]):
        if game is not None and not isinstance(game, Game):
            raise TypeError(f"Expected None or Game, got {game!r}")
        self._game = game

    @property
    def lobby(self) -> AbstractLobby[TUser]:
        return self._lobby

    @lobby.setter
    def lobby(self, lobby: AbstractLobby[TUser]):
        if not isinstance(lobby, AbstractLobby):
            raise TypeError(f"Expected AbstractLobby, got {lobby!r}")
        self._lobby = lobby

    @property
    def parser(self) -> AbstractCommandParser:
        return self._parser

    @parser.setter
    def parser(self, parser: AbstractCommandParser):
        if not isinstance(parser, AbstractCommandParser):
            raise TypeError(f"Expected AbstractCommandParser, got {parser!r}")
        self._parser = parser

    @property
    def in_game(self) -> bool:
        """True if the game is active."""
        return self._game is not None

    @property
    def in_lobby(self) -> bool:
        """True if the game is not active."""
        return not self.in_game

    def pre_check_command(self, rc: RawCommand):
        """Override for additional command checking"""

    def parse_and_run(self, source: str, obj: str) -> List[Tuple[RawCommand, Any]]:
        """Parses commands and runs them."""
        rcs = self.parser.parse(source, obj)
        res = []
        for rc in rcs:
            self.pre_check_command(rc)
            out_i = self.dispatch(rc)
            res.append((rcs, out_i))
        return res

    def find_command(self, name: str) -> CommandHandler:
        """Finds the closes command handler."""
        matcher = FuzzyMatcher(
            choices=self.registered_commands,
            score_cutoff=self.score_cutoff,
        )
        return matcher[name]

    # TODO: Syntax for default command?
    def default_command(self, source: str, name: str, *args, **kwargs) -> NoReturn:
        """Default command definition - override."""
        rc = RawCommand(source, name, args, kwargs)
        raise MafiaBadCommand(rc)

    def pre_dispatch_check(self, cmd: str, source: str = None) -> CommandHandler:
        """Gets the command handler for a given command name, with checked access.

        Checks whether the command is available for a certain source.
        If not - raises an exception.
        """
        try:
            ch: CommandHandler = self.find_command(cmd)
        except KeyError:
            # TODO: Instead return a default command? ...
            raise KeyError(f"No such command: {cmd!r}.")

        if ch.admin:
            if source not in self.lobby.admin_names:
                raise RuntimeError(f"Command {ch.name!r} requires admin privileges.")
        if self.in_lobby and not ch.lobby:
            raise RuntimeError(f"Command {ch.name!r} not available in lobby.")
        if self.in_game and not ch.game:
            raise RuntimeError(f"Command {ch.name!r} not available during game.")
        return ch

    def dispatch(self, rc: RawCommand) -> Any:
        """Calls the relevant command given by `rc`."""
        try:
            ch: CommandHandler = self.pre_dispatch_check(rc.name, source=rc.source)
        except KeyError:
            # Default command
            return self.default_command(rc.source, rc.name, *rc.args, **rc.kwargs)
        except Exception:
            raise

        # TODO: Possibly add debug output here?
        return ch.func(self, rc.source, *rc.args, **rc.kwargs)

    def user_for(self, name: str) -> Optional[TUser]:
        """Gets the user by name from the lobby."""
        return self.lobby.get(name, None)

    @command("join", lobby=True)
    def join(self, source: str):
        """Joins the game as a player."""
        user = self.user_for(source)
        if user is None:
            raise ValueError(f"No user found for {source!r}. Possibly not implemented?")
        self.lobby.add_player(source, user)  # [source] = user

    @command("in", lobby=True)
    def _join2(self, source: str):
        """Alias for 'join'."""
        return self.join(source)

    @command("leave", lobby=True)
    def leave(self, source: str):
        """Leaves the lobby."""
        user = self.user_for(source)
        if user is None:
            raise ValueError(f"No user found for {source!r}. Possibly not implemented?")
        self.lobby.remove_player(source, user)  # [source] = user

    @command("out", lobby=True)
    def _leave2(self, source: str):
        """Alias for 'leave'."""
        return self.leave(source)

    @command("force-join", admin=True, lobby=True)
    def force_join(self, source: str, *targets: List[str]):
        """Force one or more users to join the lobby."""
        users: Dict[str, TUser] = {}
        errors: List[str] = []
        for t in targets:
            user = self.user_for(t)
            if user is None:
                errors.append(t)
            else:
                users[t] = user
        if len(errors) > 0:
            raise ValueError(f"No users found for these names: {errors!r}.")
        for n, u in users.items():
            self.lobby.add_player(n, u)

    @command("kick", admin=True, lobby=True, game=True)
    def kick(self, source: str, *args, **kwargs):
        """Kick a player from the lobby or game.

        TODO: Implement!
        Will probably need change in AbstractLobby.
        Will definitely need change in Game to handle kicking/modkilling...
        """
        # TODO: Implement.

    @command("create-game", admin=True, lobby=True)
    def create_game(self, source: str, builder_name: str, *args, **kwargs):
        """Creates the game.

        TODO: Need to standardize builder options.
        For example, 'game name' would be pretty important!
        """
        if self.in_game:
            raise ValueError("BUG. Currently in a game - can't create one.")
        builder = GameBuilder.load(builder_name)
        game = builder.build(self.lobby.player_names, *args, **kwargs)
        self.game = game

    @command("destroy-game", admin=True, game=True)
    def destroy_game(self, source: str):
        """Stops the game immediately.

        Do we even need this? I assume you'd just make a new runner... :)

        TODO: Maybe a bit more elegantly?...
        TODO: Auto-stop the game when EGameEnded occurs? Not sure how that would work.
        """
        self.game = None

    @command("phase", admin=True, game=True)
    def change_phase(self, source: str, new_phase: Optional[str] = None):
        """Changes the phase (admin action)."""
        if self.game is None:
            raise RuntimeError("Game not yet started.")
        self.game.change_phase(new_phase=new_phase)
        # Alternatively:
        # e = ETryPhaseChange(self.game, new_phase=new_phase)
        # self.game.process_event(e)

    @command("do", game=True)
    def do(self, source: str, abil_name: str, *args, **kwargs):
        """Activates an ability."""
        abil_path = get_path(source, ABILITY, abil_name)
        e = EActivate(self.game, abil_path, *args, **kwargs)
        self.game.process_event(e)

    @command("vote", game=True)
    def vote(self, source: str, *args, **kwargs):
        """Votes for the target(s)."""
        return self.do(source, "vote", *args, **kwargs)

    @command("unvote", game=True)
    def unvote(self, source: str, *args, **kwargs):
        """Unvotes from all previous votes."""
        # TODO: 'vote' for "unvote-all" or smth?
        # TODO: Maybe add this only in ForumCommandRunner?
        return self.do(source, "vote", *args, **kwargs)

    # TODO: Maybe allow backup command handlers for unknown commands? :)



"""All imports for commands."""

# flake8: noqa
import warnings as _w

from .lobby import AbstractLobby, AutoAddStrLobby, SimpleDictLobby
from .parser import AbstractCommandParser, ShellCommandParser
from .raw import RawCommand, TUser
from .runner import CommandHandler, CommandRunner, command

PYGMENTS_AVAILABLE: bool
try:
    from .cli_lexer import ImplicitMafiaCliLexer, MafiaCliLexer

    PYGMENTS_AVAILABLE = True
except ImportError:
    _w.warn("Pygmants not installed, highlighting is not available.", ImportWarning)
    PYGMENTS_AVAILABLE = False



from __future__ import annotations

import inspect
import logging
from abc import ABCMeta
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    DefaultDict,
    Dict,
    Optional,
    Type,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from makefun import wraps

from open_mafia_engine.errors import MafiaAmbiguousTypeName, MafiaConverterError
from open_mafia_engine.util.classes import class_name
from open_mafia_engine.util.repr import ReprMixin

if TYPE_CHECKING:
    from open_mafia_engine.core.game import Game

NoneType = type(None)

logger = logging.getLogger(__name__)

__abstract_types__: Dict[str, Type[GameObject]] = {}
__concrete_types__: Dict[str, Type[GameObject]] = {}


class _BAD_HINT(object):
    """Sentinel for a bad type hint."""


def _get_ns() -> dict:
    from open_mafia_engine.core.all import Game

    res = locals()
    res.update(__concrete_types__)
    res.update(__abstract_types__)
    res["Game"] = Game
    return res


class GameObjectMeta(ABCMeta):
    """Metaclass for game objects."""

    def __new__(mcs, name, bases, namespace, **kwargs):
        global __abstract_types__
        global __concrete_types__

        # Modify the class as needed
        nsp = dict(namespace)
        if "__init__" in nsp:
            nsp["__init__"] = inject_converters(nsp["__init__"])

        cls = super().__new__(mcs, name, bases, nsp, **kwargs)

        # Add class to types
        cn = class_name(cls)
        existing = __abstract_types__.get(cn, __concrete_types__.get(cn))
        if existing is not None:
            logger.info(str(MafiaAmbiguousTypeName(cls, existing)))
            # raise MafiaAmbiguousTypeName(cls, existing)
            # NOTE: This + generated abilities + pickling = nightmare...
            # I guess we can assume generated abilities are the same? :)
            return existing  # just return the existing class!
        if inspect.isabstract(cls):
            __abstract_types__[cn] = cls
        else:
            __concrete_types__[cn] = cls
        return cls


class _Converter(object):
    """Global converter object."""

    def __init__(self):
        self._map: DefaultDict[
            Type[GameObject], Dict[Type, Callable[[Game, Any], GameObject]]
        ] = defaultdict(dict)

    def register(self, func: Callable) -> Callable:
        """Register this function as a converter."""

        sig = inspect.signature(func)
        type_hints = get_type_hints(func)

        params = list(sig.parameters.values())
        try:
            p_game, p_arg = params
        except Exception:
            raise TypeError(f"Expected exactly 2 arguments, but got signature: {sig!r}")

        # Check that we have everything
        try:
            return_type = type_hints["return"]
            arg_type = type_hints[p_arg.name]
            # game_type = type_hints[p_game.name]
        except KeyError as e:
            raise TypeError(f"converter() requires type annotations!") from e
        if not issubclass(return_type, GameObject):
            raise TypeError(f"Only converters for `GameObject` are allowed.")

        # Using defaultdict
        self._map[return_type][arg_type] = func

        return func

    def can_convert_to(self, type_: Type) -> bool:
        """Whether we can theoretically convert to `type_`."""

        if type_ is None:
            return True

        _t_origin = get_origin(type_)
        _t_args = get_args(type_)

        if _t_origin is Union:
            return all(self.can_convert_to(x) for x in _t_args)
        elif _t_origin is not None:
            return False

        try:
            if issubclass(type_, GameObject):
                return True
        except Exception:
            # For example, if `type_` is a Typing object...
            return False

        return False

    def convert(
        self,
        game: Game,
        type_: Union[NoneType, Type[GameObject]],
        obj: Optional[Any],
    ) -> Optional[GameObject]:
        """Convert `obj` to the given `type_`, in the context of the `game`."""

        if (type_ is None) or (type_ is NoneType):
            if obj is None:
                return None
            raise TypeError(f"obj must be None, got {obj!r}")

        if isinstance(obj, type_):  # also works if None, NoneType :)
            return obj

        _t_origin = get_origin(type_)
        _t_args = get_args(type_)

        if _t_origin is Union:  # Union or Optional
            for arg in _t_args:
                try:
                    return self.convert(game, arg, obj)
                except Exception:
                    pass
            # We went through all args - couldn't convert.
            raise MafiaConverterError(obj, type_)
        elif _t_origin is not None:
            raise TypeError(f"{_t_origin!r} is not supported for conversion.")

        # Find the function - try direct match, or using first subclass that works
        func_map = self._map[type_]
        func = func_map.get(type(obj))
        if func is not None:
            return func(game, obj)
        for t, f in func_map.items():
            if isinstance(obj, t):
                try:
                    return f(game, obj)
                except Exception:
                    logger.exception(f"Couldn't convert using {f!r}, trying again.")
        raise MafiaConverterError(obj, type_)


converter = _Converter()


def inject_converters(func: Callable) -> Callable:
    """Decorator that adds converters for all possible types."""

    if hasattr(func, "__is_converting__"):
        return func

    @wraps(func)
    def wrapper(*args, **kwargs):
        # Work with signature
        sig = inspect.signature(func)
        sb = sig.bind(*args, **kwargs)
        sb.apply_defaults()  # we want to convert the default,s too!
        # Get type hints
        # FIXME: Unsure whether this will work for external subclasses.
        type_hints = get_type_hints(func, localns=_get_ns())

        game_param = sig.parameters.get("game")

        if game_param is None:
            self_param = sig.parameters.get("self")
            if self_param is None:
                raise TypeError(f"`game` or `self` are required, got {sig!r}")
            self: GameObject = sb.arguments["self"]
            if not isinstance(self, GameObject):
                raise TypeError(f"`self` must be GameObject, got {self!r}")
            game: Game = self.game
        else:
            game: Game = sb.arguments["game"]
        params = list(sig.parameters.values())

        nargs = []
        nkw = {}
        for p in params:
            val = sb.arguments[p.name]
            if p.kind == inspect.Parameter.VAR_POSITIONAL:
                # TODO: special behavior for sequence of params
                # e.g. f(game, *actors: List[Actor])
                nargs.extend(val)
            elif p.kind == inspect.Parameter.VAR_KEYWORD:
                # TODO: special behavior for mapping of params
                # e.g. f(game, **maps: Dict[str, Actor])
                nkw.update(val)
            else:
                th = type_hints.get(p.name, _BAD_HINT)
                if converter.can_convert_to(th):
                    val = converter.convert(game, th, val)
                    # except MafiaConverterError:
                    #     # TODO: Pre-check instead?
                    #     pass

                if p.kind == inspect.Parameter.POSITIONAL_ONLY:
                    nargs.append(val)
                elif p.name in sb.kwargs:
                    nkw[p.name] = val
                else:
                    nargs.append(val)

        res = func(*nargs, **nkw)
        return res

    setattr(wrapper, "__is_converting__", True)  # FIXME
    return wrapper


class GameObject(ReprMixin, metaclass=GameObjectMeta):
    """Base class for game objects."""

    def __init__(self, game, /):
        from open_mafia_engine.core.game import Game

        if not isinstance(game, Game):
            raise TypeError(f"Expected Game, got {game!r}")
        self._game = game
        game.add(self)

    @property
    def game(self) -> Game:
        return self._game



from open_mafia_engine.core.all import Ability, Action, Actor, Game, GameObject


class DeathCausingAction(Action):
    """Action that causes an Actor to die.

    Note that KillAction and LynchAction, which are subclasses, are not subclasses
    of each other. This is to differentiate "kills" and "lynches" and other action
    types that cause death.
    """

    def __init__(
        self,
        game: Game,
        source: GameObject,
        /,
        target: Actor,
        *,
        priority: float = 0.0,
        canceled: bool = False,
    ):
        self._target = target
        super().__init__(game, source, priority=priority, canceled=canceled)

    @property
    def target(self) -> Actor:
        return self._target

    @target.setter
    def target(self, v: Actor):
        self._target = v

    def doit(self):
        self.target.status["dead"] = True


class KillAction(DeathCausingAction):
    """Action that kills the target."""


KillAbility = Ability.generate(
    KillAction,
    params=["target"],
    name="KillAbility",
    doc="Ability to perform kills",
    desc="Kills the target.",
)


class LynchAction(DeathCausingAction):
    """Action that lynches the target."""



from open_mafia_engine.core.all import (
    Outcome,
    OutcomeAction,
    OutcomeChecker,
    handler,
    EStatusChange,
)


class OCLastFactionStanding(OutcomeChecker):
    """Win if your faction is last alive. Lose if all members die."""

    @handler
    def check_deaths(self, event: EStatusChange):
        if event.key == "dead":
            own = self.parent.actors
            enemy = [x for x in self.game.actors if x not in own]
            if all(ac.status.get("dead", False) for ac in own):
                return [
                    OutcomeAction(
                        self.game, self, faction=self.parent, outcome=Outcome.defeat
                    )
                ]
            if all(ac.status.get("dead", False) for ac in enemy):
                return [
                    OutcomeAction(
                        self.game, self, faction=self.parent, outcome=Outcome.victory
                    )
                ]



try:
    import importlib.metadata as _im
except ImportError:
    import importlib_metadata as _im

from typing import Dict

import pkg_resources

__all__ = ["version", "get_versions"]
version = _im.version("open-mafia-engine")


def get_versions() -> Dict[str, str]:
    """Returns version info for all Open Mafia (sub-)packages."""

    prefix = "open-mafia"
    z = sorted(
        [
            (p.key, p.version)
            for p in pkg_resources.working_set
            if p.key.startswith(prefix)
        ]
    )
    res = {k: v for k, v in z}
    return res
